#include "bpp24.h"

int bpp24_bitblt(int x,int y,int w,int h,char *bmp)
{
  int n;
  int addr = y * swScanWidth + x * 3;
  unsigned char * gp = vga_getgraphmem();
  unsigned char * p  = gp;
  unsigned char * s  = bmp;
  int page = addr >>16;
  int next;
  int dp;
  
  vga_setpage(page);
  w *= 3;
  addr &= 0xffff;
  next = addr + w;
  p   += addr;
  
  for (n=0;n<h;n++,s+=w,addr+=swScanWidth,next+=w) {
    if ((addr&0xff0000)) {
      page++;
      vga_setpage(page);
      addr &= 0xffff;
      next = addr + w;
      p = gp;
      p += addr;
    }
    if ((addr&0xff0000)==(next&0xff0000)) {
      memcpy(p,s,w);
      p += swScanWidth;
    } else {
      dp = 0x10000 - addr;
      memcpy(p,s,dp);
      vga_setpage(page+1);
      p = gp;
      memcpy(p,s+dp,w-dp);
    }
  }
  
  return 1;
}

int bpp24_getbitmap(int x,int y,int w,int h,char *bmp)
{
  unsigned char * gp = vga_getgraphmem();
  unsigned char * p  = gp;
  int nx,ny,dw;
  int addr = y * swScanWidth + x * 3;
  int next;
  int page = addr>>16;
  
  vga_setpage(page);
  w *= 3;
  addr &= 0xffff;
  next = addr + w;
  p+=addr;

  for (ny=0;ny<h;ny++,addr+=swScanWidth,next+=swScanWidth) {
    if ((addr&0xff0000)) {
      page++;
      vga_setpage(page);
      addr &= 0xffff;
      next = addr + w;
      p = gp;
      p+= addr;
    }
    if ((addr&0xff0000)!=(next&0xff0000)) {
      dw = 0x10000 - addr;
      memcpy(bmp,p,dw);
      bmp+=dw;
      vga_setpage(page+1);
      p = gp;
      dw = w - dw;
      memcpy(bmp,p,dw);
      bmp += dw;
    } else {
      memcpy(bmp,p,w);
      p += swScanWidth; bmp += w;
    }
  }
  
  return 1;
}

int bpp24_scroll(int x,int y,int w,int h,int line)
{
  char * bmp;
  char * piece;
  int H;
  
  if (line==0) return 0;
  bmp = (char*)malloc(w*h*sizeof(char)*3);
  bpp24_getbitmap(x,y,w,h,bmp);
  if (line>0) { /* scroll up */
    piece = &bmp[line*w*4]; H = h - line;
  } else { /* scroll down */
    piece = bmp; H = h + line; y-=line;
  }
  bpp24_bitblt(x,y,w,H,piece);
  free(bmp);
  return line;
}

int bpp24_fillbox(int x,int y,int w,int h)
{
  int x2 = x + w - 1;
  int n;
  for (n=0;n<h;n++,y++) vga_drawline(x,y,x2,y);
  return 1;
}

int bpp24_xorbox(int x,int y,int w,int h)
{
  unsigned char * bmp;
  unsigned char * p;
  int n = w * h;
  bmp = (unsigned char *)malloc(n*3 + 6);
  p = bmp;
  getbitmap(x,y,w,h,bmp);
  while (n>0) { n--; *p ^= 0x00ffffff; p+=3; }
  bitblt(x,y,w,h,bmp);  
  free(bmp);
}

int bpp24_drawfont(int x,int y,int w,int h,unsigned char *bmp)
{
  char MASK[8]={0x80,0x40,0x20,0x10,0x8,0x4,0x2,0x1};
  int carry = w%8;
  int X,Y,i;
  int xp,yp;
  
  #define FG foreground
  #define BG background
  
  for (Y=0;Y<h;Y++) {
    xp = x; yp = y + Y;
    for (X=0,i=0;X<w;X++,xp++) {
      ((*bmp) & MASK[i]) ? vga_setrgbcolor(FG.r,FG.g,FG.b):
                           vga_setrgbcolor(BG.r,BG.g,BG.b);
      vga_drawpixel(xp,yp);
      if (i==7) bmp++;
      i++; i%=8;
    }
    if (carry) bmp++;
  }
  
  #undef  FG
  #undef  BG
}

int bpp24_setforeground(int r,int g,int b)
{
  set_swcolor(foreground,r,g,b);
}

int bpp24_setbackground(int r,int g,int b)
{
  set_swcolor(background,r,g,b);
}

int bpp24_miximage(unsigned char *dest,unsigned char *src,int w,int h)
{
  int n = w * h;
  while (n>0) {
    if (src[0]||src[1]||src[2]) {
      *dest=*src; dest++; src++;
      *dest=*src; dest++; src++;
      *dest=*src; dest++; src++;
    } else {
      dest++; src++;
      dest++; src++;
      dest++; src++;
    }
    n--;
  }
  return 1;
}

unsigned char * bpp24_xpm(int *w,int *h,char **xpm)
{
  int x,y;
  int W,H;
  int colors;
  int n,i;
  int present;
  unsigned char * image;
  char ** KEY;
  char * buff;
  unsigned char *p;
  swCOLOR * pal;

  if (sscanf(xpm[0],"%d %d %d %d",w,h,&colors,&present)!=4) return NULL;
  W = *w; H = *h;

  /* allocate memory */
  image = (unsigned char *)malloc(W*H*sizeof(char)*3);

  /* color */
  pal = (swCOLOR*)malloc(sizeof(swCOLOR)*colors);
  KEY = (char**)malloc(sizeof(char*)*colors);

  for (i=0,n=1;i<colors;i++,n++) {
    KEY[i]=(char*)malloc(sizeof(char)*(present+1));
    KEY[i][present]=0;
    memcpy(KEY[i],xpm[n],present);
    buff = &xpm[n][present];
    while (*buff&&*buff!='c') buff++;
    if (*buff=='c') buff++;
    if (*buff==' ') buff++;
    if (strcasecmp(buff,"None")==0) {
      memset(&pal[i],0,sizeof(swCOLOR)); 
    } else {
      if (*buff=='#') buff++;
      #define GETID(v) (v>='0'&&v<='9') ? v-'0': ((v>='A'&&v<='Z') ? v-'A'+10: ((v>='a'&&v<='z') ? v-'a'+10:0))
      #define GETCOLOR(c) c=GETID(*buff); c<<=4; buff++; c|=GETID(*buff); buff++
      GETCOLOR(pal[i].r);
      GETCOLOR(pal[i].g);
      GETCOLOR(pal[i].b);
      #undef  GETCOLOR
      #undef  GETID
    }
  }
  p = image; *p = 0;
  for (y=0;y<H;y++,n++) {
    for (x=0;x<W;x++) {
      for (i=0;i<colors;i++)
        if (strncmp(KEY[i],&xpm[n][x*present],present)==0) break;
      *p = (unsigned char)pal[i].r; p++;
      *p = (unsigned char)pal[i].g; p++;
      *p = (unsigned char)pal[i].b; p++;
    }
  }
  
  free(pal);
  for (i=0;i<colors;i++) free(KEY[i]);
  free(KEY);
  return image;
}

int bpp24_mouseinit(char **xpm)
{
  mouse_bitmap = create_image_xpm(&mouse_width,&mouse_height,xpm);
  mouse_memsize = mouse_width*mouse_height*sizeof(char)*3;
  mouse_screen = (unsigned char *)malloc(mouse_memsize);
  mouse_paint  = (unsigned char *)malloc(mouse_memsize);
  
  return 1;
}

int bpp24_mouserecover(void)
{
  if (!mouse_enable) return 0;
  if (mouse_painted) {
    bitblt(mouse_x,mouse_y,mouse_width,mouse_height,mouse_screen);
  }
  mouse_painted = 0;
  mice_update  = 0;
  return 1;
}

int bpp24_mousemove(int x,int y)
{
  if (!mouse_enable) return 0;
  if (mouse_painted) {
    bitblt(mouse_x,mouse_y,mouse_width,mouse_height,mouse_screen);
  }
  getbitmap(x,y,mouse_width,mouse_height,mouse_screen);
  memcpy(mouse_paint,mouse_screen,mouse_memsize);
  miximage(mouse_paint,mouse_bitmap,mouse_width,mouse_height);
  mouse_x = x;
  mouse_y = y;
  bitblt(x,y,mouse_width,mouse_height,mouse_paint);
  mouse_painted = 1;
  mice_update  = 0;
  return 1;
}

unsigned char * bpp24_createbuffer(int w,int h)
{
  return (unsigned char *)malloc(w*h*sizeof(char)*3);
}

int  bpp24_init(void)
{
  int n;
  drawpixel        = vga_drawpixel;
  getpixel         = vga_getpixel;
  drawline         = vga_drawline;
  setcolor         = vga_setrgbcolor;
  bitblt           = bpp24_bitblt;
  getbitmap        = bpp24_getbitmap;
  miximage         = bpp24_miximage;
  scroll           = bpp24_scroll;
  fillbox          = bpp24_fillbox;
  xorbox           = bpp24_xorbox;
  drawfont         = bpp24_drawfont;
  setforeground    = bpp24_setforeground;
  setbackground    = bpp24_setbackground;
  create_image_xpm = bpp24_xpm;
  mice_grinit      = bpp24_mouseinit;
  create_buffer    = bpp24_createbuffer;
  
  return 1;
}


